/*
 * Copyright (c) 2016-2021, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *    ======== rfAudioRx.c ========
 */

/* C standard library includes */
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>

/* Kernel Includes */
#include <ti/sysbios/knl/Event.h>
#include <ti/sysbios/BIOS.h>

/* Audio HAL top level include */
#include <ti/audio/hal/AudioHAL.h>
#include <ti/audio/stream/Packetizer.h>

/* TI-RTOS */
#include <ti/sysbios/knl/Event.h>
#include <ti/drivers/Power.h>
#include <ti/drivers/power/PowerCC26XX.h>
#include <ti/sysbios/BIOS.h>
#include <ti/drivers/dpl/HwiP.h>
#include <ti/drivers/rf/RF.h>
#include <ti/drivers/GPIO.h>

/*RF includes */
#include "RFQueue.h"
#include "smartrf_settings/smartrf_settings.h"

/* Board Header file */
#include "Board.h"

/* Audio Frame settings */
#define SAMPLE_RATE         44100
#define NUM_CHAN            AUDIOHAL_CHANNELS_MONO

#define PCM_FRAME_SIZE     60
#define ENC_FRAME_SIZE     (PCM_FRAME_SIZE * 2)

/* RTOS Event defines used to signal task from CB */
#define AUDIO_FRAME_AVAILABLE_EVENT     Event_Id_00
#define RX_EVENT                        Event_Id_01
#define ALL_EVENTS                      (AUDIO_FRAME_AVAILABLE_EVENT | RX_EVENT)

/* Events handler */
static Event_Handle eventHandle;

/* Packet RX Configuration */
#define NUM_DATA_ENTRIES       2                        /* NOTE: Only two data entries supported at the moment */
#define NUM_APPENDED_BYTES     2                        /* The Data Entries data field will contain:
                                                         * 1 Header byte (RF_cmdPropRx.rxConf.bIncludeHdr = 0x1)
                                                         * The payload bytes
                                                         * 1 status byte (RF_cmdPropRx.rxConf.bAppendStatus = 0x1) */
#define TX_FRAME_SIZE          (ENC_FRAME_SIZE + 1)     /* One byte is added to store the packet's length */
#define MAX_FRAME_QUEUE        (10)                     /* Max number of frame stored in Radio's queue*/

/* Toggle frequency of the indication LED */
#define RX_LED_TOGGLE_FREQUENCY (50)            /* The LED will blink every RX_LED_TOGGLE_FREQUENCY packets received */

/* Define radioFrame */
typedef struct radioAudioFrame_s {
    uint8_t len;
    uint8_t buffer[TX_FRAME_SIZE];
} radioAudioFrame_t;

/* This is the handle used for the AudioHAL */
static AudioHAL_Handle tlv320Handle = NULL;

/* These are the actual sample buffers that are used by the driver below the HAL */
static int16_t writeBuf[NUM_CHAN][PCM_FRAME_SIZE][AUDIOHAL_DEFAULT_BUF_DEPTH];

/* Single frame buffer used to copy data in from HAL to application space */
static int16_t audioOut[NUM_CHAN][PCM_FRAME_SIZE];

/* Static buffer of MAX_FRAME_QUEUE queue elements for radio transmission */
static radioAudioFrame_t radioRxFrameQueue[MAX_FRAME_QUEUE] = {0};
static uint8_t rxFramesTail = 0U;
static uint8_t rxFramesHead = 0U;

/* Radio handles */
static RF_Object rfObject;
static RF_Handle rfHandle;

/* Radio packet buffer */
static uint8_t rxDataEntryBuffer[RF_QUEUE_DATA_ENTRY_BUFFER_SIZE(NUM_DATA_ENTRIES, TX_FRAME_SIZE, NUM_APPENDED_BYTES)];

/* Receive dataQueue for RF Core to fill in data */
static dataQueue_t rxDataQueue;
static rfc_dataEntryGeneral_t* currentDataEntry;

/* Packets counters */
uint32_t numberOfRejectedFrames = 0U; //Only used if the Radio frames have always the same size
uint32_t numberOfReceivedPackets = 0U;
uint32_t numberOfMissedPackets = 0U;
uint32_t numberOfCorruptedPackets = 0U;
uint32_t numberOfCorrectPackets = 0U;

/*
 *  ======== enqueueRxFrame ========
 *  Enqueue an audio frame into the radio frame queue
 */
static void enqueueRxFrame(uint8_t* buf, uint8_t size)
{
    /* Critical section to ensure that no modification of the list has been done concurrently */
    uint32_t key = HwiP_disable();

    if((size!=0) && (buf!=NULL)){

        if(size==ENC_FRAME_SIZE)
        {                          // Only the incoming frames with the proper size are kept
                                   // (should be done only if the frame size is known and fixed)

            /* ring buffer style! */
            memcpy(&radioRxFrameQueue[rxFramesHead].buffer, buf, size);
            radioRxFrameQueue[rxFramesHead++].len = size;

            /* radioRxFrameQueue is a ring-list */
            if (rxFramesHead == MAX_FRAME_QUEUE)
            {
                rxFramesHead = 0;
            }

            /* Release critical section */
            HwiP_restore(key);

            Event_post(eventHandle, AUDIO_FRAME_AVAILABLE_EVENT);

        }
        else { // Reject frames without the proper size
            numberOfRejectedFrames ++;
        }
    }
}

/*
 *  ======== getRxFrame ========
 *  Get an audio frame from the radio frame queue
 */
static void getRxFrame(uint8_t* buf, uint32_t* size) {

    /* In critical section ... */
    uint32_t key = HwiP_disable();

    /* ring buffer style! */
    if (rxFramesTail != rxFramesHead) {
        memcpy(buf, &radioRxFrameQueue[rxFramesTail].buffer, radioRxFrameQueue[rxFramesTail].len);
        *size = radioRxFrameQueue[rxFramesTail++].len;

        if (rxFramesTail == MAX_FRAME_QUEUE) {
            rxFramesTail = 0;
        }
    }
    else
    {
        /* Cannot find a buffer: the list is empty */
        *size = 0;
    }

    /* Release critical section */
    HwiP_restore(key);
}

/*
 *  ======== rxDoneCB ========
 *  Callback function executed when an RF command is finished
 */
void rxDoneCB(RF_Handle h, RF_CmdHandle ch, RF_EventMask e)
{
    AudioHAL_onPacketRxDone(tlv320Handle, PCM_FRAME_SIZE);

    if (e & RF_EventRxEntryDone)
    {
        /* Get current unhandled data entry */
        currentDataEntry = RFQueue_getDataEntry();

        /* Handle the packet data, located at &currentDataEntry->data:
         * - Length is the first byte with the current configuration
         * - Data starts from the second byte */
        uint8_t* packetDataPointer = (&currentDataEntry->data + 1);

        /* Put the packet into the queue */
        enqueueRxFrame(&packetDataPointer[1], packetDataPointer[0]);

        Event_post(eventHandle, RX_EVENT);

        RFQueue_nextEntry();
    }
}

/*
 *  ======== audioDataNeededCB ========
 *  Callback function executed when audio data are expected by the AudioHAL
 */
void audioDataNeededCB(AudioHAL_Handle handle, int_fast16_t status)
{
    /* Nothing to do */
}

/*
 *  ======== audioEventCB ========
 *  Callback function executed when an audio event occurs
 */
void audioEventCB(AudioHAL_Handle handle, AudioHALEvent event)
{
    /* Nothing to do */
}

/*
 *  ======== mainThread ========
 */
void *mainThread(void *arg0)
{
    /* Audio HAL related variables */
    AudioHAL_Status stat;
    AudioHAL_Params audioHalParams;

    /* Events related variable */
    uint32_t events;

    /* Call driver init functions */
    Board_initGeneral();
    AudioHAL_init();
    Packetizer_init();

    /* Create event used to signal frame processing */
    eventHandle = Event_create(NULL, NULL);
    if(eventHandle == NULL)
    {
        /* Event_create(NULL, NULL) failed */
        while (1);
    }

    /* Setup radio */
    RF_Params rfParams;
    RF_Params_init(&rfParams);

    if( RFQueue_defineQueue(&rxDataQueue,
                            rxDataEntryBuffer,
                            sizeof(rxDataEntryBuffer),
                            NUM_DATA_ENTRIES,
                            TX_FRAME_SIZE + NUM_APPENDED_BYTES))
    {
        /* Failed to allocate space for all data entries */
        while(1);
    }

    /* Modify CMD_PROP_RX command for application needs */
    /* Set the Data Entity queue for received data */
    RF_cmdPropRx.pQueue = &rxDataQueue;
    /* Discard ignored packets from Rx queue */
    RF_cmdPropRx.rxConf.bAutoFlushIgnored = 1;
    /* Discard packets with CRC error from Rx queue */
    RF_cmdPropRx.rxConf.bAutoFlushCrcErr = 1;
    /* Implement packet length filtering to avoid PROP_ERROR_RXBUF */
    RF_cmdPropRx.maxPktLen = TX_FRAME_SIZE;
    RF_cmdPropRx.pktConf.bRepeatOk = 1;
    RF_cmdPropRx.pktConf.bRepeatNok = 1;

    /* Request access to the radio */
    rfHandle = RF_open(&rfObject, &RF_prop,
                       (RF_RadioSetup*)&RF_cmdPropRadioDivSetup, &rfParams);

    RF_runCmd(rfHandle, (RF_Op*)&RF_cmdFs, RF_PriorityNormal, NULL, 0);

    /* Set AudioHAL settings */
    AudioHAL_Params_init(&audioHalParams);
    audioHalParams.frameSize = PCM_FRAME_SIZE;
    audioHalParams.numChannels = NUM_CHAN;
    audioHalParams.writeBuf = writeBuf;
    audioHalParams.bitDepth = AUDIOHAL_BITDEPTH_16;
    audioHalParams.samplingFrequency = SAMPLE_RATE;
    audioHalParams.readBufDepth = AUDIOHAL_DEFAULT_BUF_DEPTH;
    audioHalParams.writeBufDepth = AUDIOHAL_DEFAULT_BUF_DEPTH;
    audioHalParams.dataNeededCB = audioDataNeededCB;
    audioHalParams.eventCB = audioEventCB;
    audioHalParams.direction = AUDIOHAL_OUTPUT;

    audioHalParams.role = AUDIOHAL_SLAVE; // Mandatory for WCD - Make sure you have properly configured the boosterpack

    /* Open AudioHAL interface to TLV320 */
    tlv320Handle = AudioHAL_open(Board_AUDIOHAL_CODEC, &audioHalParams);

    if(tlv320Handle == NULL)
    {
        /* AudioHAL_open(Board_AUDIOHAL_CODEC, ...) failed */
        while (1);
    }

    stat = AudioHAL_startStream(tlv320Handle, AUDIOHAL_OUTPUT);

    if(stat != AUDIOHAL_SUCCESS)
    {
        /* AudioHAL_startStream(...) failed */
        while (1);
    }

    /* Start infinite RX */
    RF_CmdHandle cmdHandle = RF_postCmd(rfHandle, (RF_Op*)&RF_cmdPropRx,
                                                   RF_PriorityNormal, &rxDoneCB,
                                                   RF_EventRxEntryDone);

    while(1)
    {
        events = Event_pend(eventHandle, Event_Id_NONE, ALL_EVENTS, BIOS_WAIT_FOREVER);

        if(events)
        {

            if(events & AUDIO_FRAME_AVAILABLE_EVENT)
            {
                uint32_t decLen;

                getRxFrame((uint8_t*)audioOut, &decLen);

                /* Place buffer in the output */
               AudioHAL_writeBufPut(tlv320Handle, audioOut);

            }

            if(events & RX_EVENT)
            {
                numberOfReceivedPackets ++;
                if(numberOfReceivedPackets % RX_LED_TOGGLE_FREQUENCY == 0){
                    GPIO_toggle(Board_GPIO_LED0);
                }
            }
        }
    }
}
