/*
 * Copyright (c) 2016-2021, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *    ======== rfAudioTx.c ========
 */

/* C standard library includes */
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>

/* Kernel Includes */
#include <ti/sysbios/knl/Event.h>
#include <ti/sysbios/BIOS.h>

/* Audio HAL top level include */
#include <ti/audio/hal/AudioHAL.h>
#include <ti/audio/stream/Packetizer.h>

/* TI-RTOS */
#include <ti/sysbios/knl/Event.h>
#include <ti/drivers/Power.h>
#include <ti/drivers/power/PowerCC26XX.h>
#include <ti/sysbios/BIOS.h>
#include <ti/drivers/dpl/HwiP.h>
#include <ti/drivers/rf/RF.h>
#include <ti/drivers/GPIO.h>

/*RF includes */
#include "smartrf_settings/smartrf_settings.h"

/* Board Header file */
#include "Board.h"

/* Audio Frame settings (i.e. frames read from the AudioHaAL) */
#define SAMPLE_RATE         44100
#define NUM_CHAN            AUDIOHAL_CHANNELS_MONO

#define PCM_FRAME_SIZE     60
#define ENC_FRAME_SIZE     (PCM_FRAME_SIZE * 2)


/* RTOS Event defines used to signal task from CB */
#define AUDIO_DATA_READY_EVENT          Event_Id_00
#define AUDIO_READY_TO_SEND_EVENT       Event_Id_01
#define RF_TX_DONE_EVENT                Event_Id_02
#define ALL_EVENTS                      (AUDIO_DATA_READY_EVENT | AUDIO_READY_TO_SEND_EVENT | RF_TX_DONE_EVENT)

/* Events handler */
static Event_Handle eventHandle;

/* Packet RX Configuration */
#define TX_FRAME_SIZE          (ENC_FRAME_SIZE + 1)     /* One byte is added to store the packet's length */
#define MAX_FRAME_QUEUE        (2)                     /* Max number of frame stored in Radio's queue*/

/* Toggle frequency of the indication LED */
#define TX_LED_TOGGLE_FREQUENCY (50)                    /* The LED will blink every TX_LED_TOGGLE_FREQUENCY packets sent */

/*Define radioFrame */
typedef struct radioAudioFrame_s {
    uint8_t len;
    uint8_t buffer[TX_FRAME_SIZE];
} radioAudioFrame_t;

/* These are the actual sample buffers that are used by the driver below the HAL */
static int16_t readBuf[NUM_CHAN][PCM_FRAME_SIZE][AUDIOHAL_DEFAULT_BUF_DEPTH];

#ifdef ENABLE_ECHO
static int16_t writeBuf[NUM_CHAN][PCM_FRAME_SIZE][AUDIOHAL_DEFAULT_BUF_DEPTH];
#endif // ENABLE_ECHO

/* Single frame buffer used to copy data in from HAL to application space */
static int16_t audioIn[NUM_CHAN][PCM_FRAME_SIZE];

/* Static buffer of MAX_FRAME_QUEUE queue elements for radio transmission */
static radioAudioFrame_t radioTxFrameQueue[MAX_FRAME_QUEUE] = {0};
static uint8_t txFramesTail = 0;
static uint8_t txFramesHead = 0;

/* Radio handles */
static RF_Object rfObject;
static RF_Handle rfHandle;
static bool radioActive = FALSE;

/* TX Counter */
uint32_t numberOfSentPackets = 0U;

/*
 *  ======== enqueueTxFrame ========
 *  Enqueue an audio frame into the radio frame queue
 */
static void enqueueTxFrame(uint8_t* buf, uint32_t size)
{
    /* In critical section ... */
    uint32_t key = HwiP_disable();
    /* ring buffer style! */
    memcpy(&radioTxFrameQueue[txFramesHead].buffer, buf, size);
    radioTxFrameQueue[txFramesHead++].len = size;

    if (txFramesHead == MAX_FRAME_QUEUE) {
        txFramesHead = 0;
    }

    /* Release critical section */
    HwiP_restore(key);

    Event_post(eventHandle, AUDIO_READY_TO_SEND_EVENT);
}

/*
 *  ======== getTxFrame ========
 *  Get an audio frame from the radio frame queue
 */
static void getTxFrame(uint8_t* buf, uint8_t* size) {

    /* In critical section ... */
    uint32_t key = HwiP_disable();

    /* ring buffer style! */
    if (txFramesTail != txFramesHead) {
        memcpy(buf, &radioTxFrameQueue[txFramesTail].buffer, radioTxFrameQueue[txFramesTail].len);
        *size = radioTxFrameQueue[txFramesTail++].len;

        if (txFramesTail == MAX_FRAME_QUEUE) {
            txFramesTail = 0;
        }
    }
    else
    {
        *size = 0;
    }

    /* Release critical section */
    HwiP_restore(key);
}

/*
 *  ======== audioEventCB ========
 *  Callback function executed when an audio event occurs
 */
void audioEventCB(AudioHAL_Handle handle, AudioHALEvent event)
{
    /* Nothing to do */
}

/*
 *  ======== audioDataReadyCB ========
 *  Callback function executed some audio data has been read
 */
void audioDataReadyCB(AudioHAL_Handle handle, int_fast16_t status)
{
    Event_post(eventHandle, AUDIO_DATA_READY_EVENT);
}



void audioDataNeeded(AudioHAL_Handle handle, int_fast16_t status)
{
    /*  Intentionally not used */
}


 /*
  *  ======== txDoneCB ========
  *  Callback function executed when an RF command is finished
  */
void txDoneCB(RF_Handle h, RF_CmdHandle ch, RF_EventMask e)
{
    if (e & RF_EventLastCmdDone)
    {
        /* We are done with this frame */
        radioActive = FALSE;
        Event_post(eventHandle, RF_TX_DONE_EVENT);
    }
    else
    {
        /* This was unexpected */
        while (1);
    }
}

/*
 *  ======== mainThread ========
 */
void *mainThread(void *arg0)
{
    /* Audio HAL related variables */
    AudioHAL_Status stat;
    AudioHAL_Params audioHalParams;
    AudioHAL_Handle tlv320Handle = NULL;

    /* Events related variable */
    uint32_t events;

    /* Call driver init functions */
    Board_initGeneral();
    AudioHAL_init();
    Packetizer_init();

    /* Create event used to signal frame processing */
    eventHandle = Event_create(NULL, NULL);
    if(eventHandle == NULL)
    {
        /* Event_create(NULL, NULL) failed */
        while (1);
    }

    /* Setup radio */
    RF_Params rfParams;
    RF_Params_init(&rfParams);


    /* Request access to the radio */
    rfHandle = RF_open(&rfObject, &RF_prop,
                       (RF_RadioSetup*)&RF_cmdPropRadioDivSetup, &rfParams);

    RF_runCmd(rfHandle, (RF_Op*)&RF_cmdFs, RF_PriorityNormal, NULL, 0);

    /* Default scheduling params */
    RF_ScheduleCmdParams    schParams;
    RF_ScheduleCmdParams_init(&schParams);

    /* Set AudioHAL settings */
    AudioHAL_Params_init(&audioHalParams);
    audioHalParams.frameSize = PCM_FRAME_SIZE;
    audioHalParams.numChannels = NUM_CHAN;
    audioHalParams.readBuf = readBuf;
    audioHalParams.bitDepth = AUDIOHAL_BITDEPTH_16;
    audioHalParams.samplingFrequency = SAMPLE_RATE;
    audioHalParams.readBufDepth = AUDIOHAL_DEFAULT_BUF_DEPTH;
    audioHalParams.dataReadyCB = audioDataReadyCB;
    audioHalParams.eventCB = audioEventCB;
    audioHalParams.direction = AUDIOHAL_INPUT;

#ifdef ENABLE_ECHO
    audioHalParams.writeBufDepth = AUDIOHAL_DEFAULT_BUF_DEPTH;
    audioHalParams.dataNeededCB = &audioDataNeeded;
    audioHalParams.writeBuf = writeBuf;
    audioHalParams.direction = AUDIOHAL_INPUT_OUTPUT;
#endif // ENABLE_ECHO

    audioHalParams.role = AUDIOHAL_SLAVE; // This is suggested when using WCD on the Rx side but not mandatory - Make sure the boosterpack is properly configured

    /* Open AudioHAL interface to TLV320 */
    tlv320Handle = AudioHAL_open(Board_AUDIOHAL_CODEC, &audioHalParams);

    if(tlv320Handle == NULL)
    {
        /* AudioHAL_open(Board_AUDIOHAL_CODEC, ...) failed */
        while (1);
    }


    stat = AudioHAL_startStream(tlv320Handle, AUDIOHAL_INPUT);

#ifdef ENABLE_ECHO
    stat |= AudioHAL_startStream(tlv320Handle, AUDIOHAL_OUTPUT);
#endif //ENABLE_ECHO

    if(stat != AUDIOHAL_SUCCESS)
    {
        /* AudioHAL_startStream(...) failed */
        while (1);
    }

    while(1)
    {
        events = Event_pend(eventHandle, Event_Id_NONE, ALL_EVENTS, BIOS_WAIT_FOREVER);

        if(events)
        {

            if(events & AUDIO_DATA_READY_EVENT)
            {
                /* Get a frame from the input buffers */
                AudioHAL_readBufGet(tlv320Handle, (void *)audioIn);

                /* Place decoded buffer in the output */
                AudioHAL_writeBufPut(tlv320Handle, (void *)audioIn);

                /* Enqueue for transmission over air */
                enqueueTxFrame((uint8_t*)audioIn, ENC_FRAME_SIZE);

            }

            if (events & AUDIO_READY_TO_SEND_EVENT)
            {

                /* Are we already in TX? If that is the case just skip ahead */
                if (radioActive == FALSE)
                {
                    /* RF packet to send out */
                    uint8_t packet[TX_FRAME_SIZE];

                    /* Try to get a frame from the queue */
                    getTxFrame(&packet[1],&packet[0]);


                    /* If size is larger then 0, we got a frame */
                    if (packet[0] > 0) {

                        /* Configure the radio packet */
                        RF_cmdPropTx.pktLen = TX_FRAME_SIZE;
                        RF_cmdPropTx.pPkt = packet;
                        RF_cmdPropTx.startTrigger.triggerType = TRIG_NOW;

                        RF_CmdHandle cmdHandle = RF_scheduleCmd(rfHandle, (RF_Op*)&RF_cmdPropTx, &schParams, txDoneCB, NULL);

                        if (cmdHandle == RF_ALLOC_ERROR) {
                           while (1);
                        }

                        radioActive = TRUE;
                   }
                }
            }

            if(events & RF_TX_DONE_EVENT){

                /* One frame has been sent out */
                numberOfSentPackets++;
                if (!(numberOfSentPackets % TX_LED_TOGGLE_FREQUENCY)){
                    GPIO_toggle(Board_GPIO_LED1);
                }
            }
        }
    }
}
